#include <lpc214x.h>

//
//  This is where the various LCD control pins are assigned.  Currently they're
//  all assigned on port 1.  If you have to break them up, you'll need to edit
//  the functions them.  It's very convenient to have the D7..D0 lines grouped
//  together.  Otherwise you'll have to bit test each bit and assign the pin.
//  Much slower.
//
#define LCD_D4        GPIO_IO_P28
#define LCD_D5        GPIO_IO_P29
#define LCD_D6        GPIO_IO_P30
#define LCD_D7        GPIO_IO_P31
#define LCD_DATA      (LCD_D7 | LCD_D6 | LCD_D5 | LCD_D4)
#define LCD_Dx_SHIFT  ((unsigned int) 28)
#define LCD_RS        GPIO_IO_P25
#define LCD_RW        GPIO_IO_P26
#define LCD_E         GPIO_IO_P27

//
//
//
#define LCD_COMMAND_CLEAR       0x01
#define LCD_COMMAND_HOME        0x02
#define LCD_COMMAND_MODE        0x04
#define LCD_COMMAND_DISPLAY     0x08
#define LCD_COMMAND_SHIFT       0x10
#define LCD_COMMAND_FUNCTION    0x20
#define LCD_COMMAND_CGRAM       0x40
#define LCD_COMMAND_DDRAM       0x80

#define LCD_MODE_SHIFT_NO       0x00
#define LCD_MODE_SHIFT_YES      0x01
#define LCD_MODE_DECREMENT      0x00
#define LCD_MODE_INCREMENT      0x02
#define LCD_MODE_MASK           0x03

#define LCD_SHIFT_CURSOR        0x00
#define LCD_SHIFT_DISPLAY       0x04
#define LCD_SHIFT_LEFT          0x00
#define LCD_SHIFT_RIGHT         0x08
#define LCD_SHIFT_MASK          0x0c

#define LCD_FUNCTION_FONT_5X8   0x00
#define LCD_FUNCTION_FONT_5X11  0x04
#define LCD_FUNCTION_LINES_1    0x00
#define LCD_FUNCTION_LINES_2    0x08
#define LCD_FUNCTION_BUS_4      0x00
#define LCD_FUNCTION_BUS_8      0x10
#define LCD_FUNCTION_MASK       0x1c

#define LCD_DISPLAY_BLINK_OFF   0x00
#define LCD_DISPLAY_BLINK_ON    0x01
#define LCD_DISPLAY_CURSOR_OFF  0x00
#define LCD_DISPLAY_CURSOR_ON   0x02
#define LCD_DISPLAY_DISPLAY_OFF 0x00
#define LCD_DISPLAY_DISPLAY_ON  0x04
#define LCD_DISPLAY_MASK        0x07

#define LCD_PWM_FREQ 20000

//
//
//  Structure for passing messages to the LCD
//
typedef struct
{
	const char *pcMessage;
}
xLCDMessage_t;

//


//
//  Each operation is 83.33ns.  Cumulative delays make be longer, but not shorter.
//  This delay assumes a 48Mhz clock and MAM=partial
//
#define delay166ns() do { asm volatile (" push {r0}" : /* no output */ : /* no inputs */ ); \
	asm volatile (" pop  {r0}" : /* no output */ : /* no inputs */ ); \
	} while (0)

//
//  166ns measured
//
static inline void lcdActiveDelay140ns (void)
{
	delay166ns ();
}

//
//  333ns measured
//
static inline void lcdActiveDelay320ns (void)
{
	delay166ns ();
	delay166ns ();
}

//
//  500ns measured
//
static inline void lcdActiveDelay450ns (void)
{
	delay166ns ();
	delay166ns ();
	delay166ns ();
}

//
//  666ns measured
//
static inline void lcdActiveDelay550ns (void)
{
	delay166ns ();
	delay166ns ();
	delay166ns ();
	delay166ns ();
}

static void lcdActiveDelayMs (unsigned int timeInMs)
{
	unsigned int microSeconds;

	while (timeInMs > 0) {
		for (microSeconds = 1000; microSeconds > 0; microSeconds--) {
		}
		timeInMs--;
	}
}

static inline void lcdDataToPort (unsigned char c)
{
	GPIO1_FIOCLR = (~c & 0x0f) << LCD_Dx_SHIFT;
	GPIO1_FIOSET = ( c & 0x0f) << LCD_Dx_SHIFT;
}

static inline unsigned int lcdDataFromPort (void)
{
	return (GPIO1_FIOPIN & LCD_DATA) >> LCD_Dx_SHIFT;
}

static inline void lcdSelectCommandRegister (void)
{
	GPIO1_FIOCLR = LCD_RS;
}

static inline void lcdSelectDataRegister (void)
{
	GPIO1_FIOSET = LCD_RS;
}

static inline void lcdSelectWriteMode (void)
{
	GPIO1_FIOCLR = LCD_RW;
}

static inline void lcdSelectReadMode (void)
{
	GPIO1_FIOSET = LCD_RW;
}

static inline void lcdEnableHigh (void)
{
	GPIO1_FIOSET = LCD_E;
}

static inline void lcdEnableLow (void)
{
	GPIO1_FIOCLR = LCD_E;
}

static inline void lcdDataAsInputs (void)
{
	GPIO1_FIODIR &= ~LCD_DATA;
}

static inline void lcdDataAsOutputs (void)
{
	GPIO1_FIODIR |= LCD_DATA;
}

//
//
//
static void lcdWritePort4 (unsigned char c)
{
	lcdEnableHigh ();
	lcdDataToPort (c & 0x0f);
	lcdActiveDelay140ns ();
	lcdEnableLow ();
	lcdActiveDelay550ns ();
	lcdActiveDelay550ns ();
}

static void lcdWriteData (unsigned char c)
{
	lcdSelectDataRegister ();
	lcdSelectWriteMode ();

	lcdWritePort4 (c >> 4);
	lcdWritePort4 (c & 0x0f);
}

static void lcdWriteCommand4 (unsigned char c)
{
	lcdSelectCommandRegister ();
	lcdSelectWriteMode ();

	lcdWritePort4 (c);
}

static void lcdWriteCommand (unsigned char c)
{
	lcdSelectCommandRegister ();
	lcdSelectWriteMode ();

	lcdWritePort4 (c >> 4);
	lcdWritePort4 (c & 0x0f);
}

static int lcdReadStatus4 (void)
{
	unsigned char c;

	lcdEnableHigh ();
	lcdDataAsInputs ();
	lcdActiveDelay140ns ();
	c = lcdDataFromPort ();
	lcdEnableLow ();
	lcdActiveDelay550ns ();
	lcdActiveDelay550ns ();

	return c & 0x0f;
}

static int lcdReadStatus (void)
{
	unsigned char c;

	lcdSelectCommandRegister ();
	lcdSelectReadMode ();

	c  = lcdReadStatus4 () << 4;
	c |= lcdReadStatus4 ();

	lcdSelectWriteMode ();
	lcdDataAsOutputs ();

	return c;
}

static void lcdWaitWhileBusy (void)
{
	while (lcdReadStatus () & 0x80)
		;
}

//
//
//
static void lcdDisplay (unsigned char display)
{
	lcdWaitWhileBusy ();
	lcdWriteCommand (LCD_COMMAND_DISPLAY | (display & LCD_DISPLAY_MASK));
}

static void lcdMode (unsigned char mode)
{
	lcdWaitWhileBusy ();
	lcdWriteCommand (LCD_COMMAND_MODE | (mode & LCD_MODE_MASK));
}

static void lcdClear (void)
{
	lcdWaitWhileBusy ();
	lcdWriteCommand (LCD_COMMAND_CLEAR);
}

static void lcdHome (void)
{
	lcdWaitWhileBusy ();
	lcdWriteCommand (LCD_COMMAND_HOME);
}

static void lcdGotoXY (unsigned int x, unsigned int y)
{
	lcdWaitWhileBusy ();
	lcdWriteCommand (LCD_COMMAND_DDRAM | (y ? 0x40 : 0x00) | (x & 0x0f));
	lcdWaitWhileBusy ();
}

static void lcdPutChar (char c)
{
	lcdWaitWhileBusy ();
	lcdWriteData (c);
}

static void lcdWriteStringCtl (const char *s)
{
	while (*s)
	{
		char c = *s++;

		switch (c)
		{
		case '\f' :
			lcdClear ();
			break;

		case 0x1e :
			lcdHome ();
			break;

		case 0x1b :
			switch (*s++)
			{
			case '=' :
				{
					unsigned char x = *s++;
					unsigned char y = *s++;

					lcdGotoXY (x, y);
				}
				break;

			case '*' :
				lcdClear ();
				break;

			case '.' :
				switch (*s++)
				{
				case '0' :
					lcdDisplay (LCD_DISPLAY_DISPLAY_ON | LCD_DISPLAY_CURSOR_OFF | LCD_DISPLAY_BLINK_OFF);
					break;

				case '1' :
					lcdDisplay (LCD_DISPLAY_DISPLAY_ON | LCD_DISPLAY_CURSOR_ON | LCD_DISPLAY_BLINK_ON);
					break;

				case '2' :
					lcdDisplay (LCD_DISPLAY_DISPLAY_ON | LCD_DISPLAY_CURSOR_ON | LCD_DISPLAY_BLINK_OFF);
					break;
				}
			}
			break;

		default :
			lcdPutChar (c);
			break;
		}
	}
}


// -------------------------------------------------------------------------------------------------------------
// Business & control interface
// -------------------------------------------------------------------------------------------------------------

void lcd_controller_init() {
	GPIO1_FIODIR |= (LCD_E | LCD_RS | LCD_RW);

	lcdEnableLow ();
	lcdSelectWriteMode ();
	lcdSelectCommandRegister ();
	lcdDataAsOutputs ();
	lcdDataToPort (0xff);

	//
	//  Queue used to send LCD task messages, only 1 outstanding message
	//  (so static buffers, such as those used in lcdCursorAt() don't
	//  get lost)

	lcdWriteCommand4 ((LCD_COMMAND_FUNCTION | LCD_FUNCTION_BUS_4) >> 4);  lcdActiveDelayMs (30);

	lcdWriteCommand4 ((LCD_COMMAND_FUNCTION | LCD_FUNCTION_BUS_4) >> 4);  lcdActiveDelayMs (30);
	lcdWriteCommand4 ((LCD_COMMAND_FUNCTION | LCD_FUNCTION_BUS_4) >> 4);  lcdActiveDelayMs (30);

	lcdWriteCommand (LCD_COMMAND_FUNCTION | LCD_FUNCTION_BUS_4 | LCD_FUNCTION_LINES_2 | LCD_FUNCTION_FONT_5X8);  lcdActiveDelayMs (30);
	lcdWriteCommand (LCD_COMMAND_FUNCTION | LCD_FUNCTION_BUS_4 | LCD_FUNCTION_LINES_2 | LCD_FUNCTION_FONT_5X8);  lcdActiveDelayMs (30);

	lcdDisplay (LCD_DISPLAY_DISPLAY_ON | LCD_DISPLAY_CURSOR_OFF | LCD_DISPLAY_BLINK_OFF);

	lcdClear ();
	lcdMode (LCD_MODE_INCREMENT | LCD_MODE_SHIFT_NO);
	lcdGotoXY (0, 0);
}

void lcd_controller_lcd_gotoXY(int x, int y) {
	static char buffer [5];

	buffer [0] = '\033';
	buffer [1] = '=';
	buffer [2] = x;
	buffer [3] = y;
	buffer [4] = '\0';

	lcd_controller_lcd_write(buffer);
}

void lcd_controller_lcd_write(const char *s) {
	lcdWriteStringCtl(s);
}


void Initialize(void)  {
	//  Use fast I/O on both ports
	SCB_SCS = SCB_SCS_GPIO0M | SCB_SCS_GPIO1M;

	//  Configure pin functions.  All pins are set to GPIO, including the Debug
	//  port (P1.26) and the Trace port (P1.25..P1.16).
	PCB_PINSEL0 = PCB_PINSEL0_ALL_GPIO;
	PCB_PINSEL1 = PCB_PINSEL1_ALL_GPIO;
	PCB_PINSEL2 = (PCB_PINSEL2_P13626_GPIO | PCB_PINSEL2_P12516_GPIO);

	//  Set all GPIO to input (safer than an output, which may be driving a high
	//  into a closed switch such as BSL).
	GPIO0_FIODIR = ~GPIO_IO_ALL;
	GPIO1_FIODIR = ~GPIO_IO_JTAG;

	//  Setup the PLL to multiply the 12Mhz XTAL input by 4, divide by 1
	SCB_PLLCFG = (SCB_PLLCFG_MUL4 | SCB_PLLCFG_DIV1);

	//  Activate the PLL by turning it on then feeding the correct sequence of bytes
	SCB_PLLCON  = SCB_PLLCON_PLLE;
	SCB_PLLFEED = SCB_PLLFEED_FEED1;
	SCB_PLLFEED = SCB_PLLFEED_FEED2;

	//  Wait for the PLL to lock...
	while (!(SCB_PLLSTAT & SCB_PLLSTAT_PLOCK));

	//  ...before connecting it using the feed sequence again
	SCB_PLLCON  = SCB_PLLCON_PLLC | SCB_PLLCON_PLLE;
	SCB_PLLFEED = SCB_PLLFEED_FEED1;
	SCB_PLLFEED = SCB_PLLFEED_FEED2;

	//  Setup and turn on the MAM.  Three cycle access is used due to the fast
	//  PLL used.  It is possible faster overall performance could be obtained by
	//  tuning the MAM and PLL settings.
	MAM_TIM = MAM_TIM_3;
	MAM_CR = MAM_CR_FULL;

	//  Setup the peripheral bus to be the same as the PLL output (48Mhz)
	SCB_VPBDIV = SCB_VPBDIV_100;

	//  Disable power to all modules
	SCB_PCONP = SCB_PCONP_ALLOFF;

	//  Make sure all interrupts disabled
	VIC_IntEnClr = VIC_IntEnClr_MASK;

	//  Lastly, switch interrupt handlers to RAM vectors
	SCB_MEMMAP = SCB_MEMMAP_URM;
}


int main (void){
    Initialize();
  
    char buffer[30];
    buffer[0] = 0x1b;
    buffer[1] = '=';
    buffer[2] = 0;
    buffer[3] = 0;
    buffer[4] =  'A';
    buffer[5] =  'H';
    buffer[6] =  'O';
    buffer[7] =  'J';
    buffer[8] =  ' ';
    buffer[9] =  'P';
    buffer[10] = 'E';
    buffer[11] = 'T';
    buffer[12] = 'R';
    buffer[13] = 'E';
    buffer[14] = 0;
    
    lcd_controller_init();
    lcdActiveDelay450ns();
    lcdActiveDelay140ns();
    lcdActiveDelay450ns();
    lcdWriteStringCtl(buffer);
}
